import 'dart:async';
import 'dart:math' as math;
import 'dart:io';

import 'package:flutter/material.dart';
import 'package:image_picker/image_picker.dart';

class Taquin extends StatefulWidget {
  final int difficulty;
  final int colNb;

  const Taquin({required this.difficulty, required this.colNb, super.key});

  @override
  State<Taquin> createState() => _Taquin();
}

class _Taquin extends State<Taquin> {
  File? _image;
  final picker = ImagePicker();

  List<Widget> tiles = <Widget>[];
  List<Key> initialOrder = <Key>[];
  double? size = 300;
  int emptyTileIndex = 0;

  late int score;
  late String imageUrl;
  late Image image;
  late int colNb;
  late int difficulty;
  Duration stopwatch = const Duration();
  Timer? stopwatchTimer;
  

  @override
  void initState(){
    super.initState();

    image = Image.network('https://picsum.photos/512');
    
    colNb = widget.colNb;
    difficulty = widget.difficulty;
    score = difficulty * colNb ;
    
    tiles = buildImage();
    moveTiles();
  }

  List<Widget> buildImage(){
    var res = <Widget>[];
    initialOrder = <Key>[];

    emptyTileIndex = 0;

    for(int i = 0; i < colNb; i++){
      for(int j = 0; j < colNb; j++){
        Tile tile = Tile(image: image, alignment: 
          Alignment(-1 + 2 * j.toDouble() / (colNb.toDouble() - 1),
                    -1 + 2 * i.toDouble() / (colNb.toDouble() - 1))
        );

        Key k = UniqueKey();
        initialOrder.add(k);

        res.add(
          Flexible(key: k, child: Container(
            margin: const EdgeInsets.all(2),
            child: Material(
              child: InkWell(
                child: tile.croppedImageTile(1 / colNb),
                onTap: () {
                  setState(() {
                    swapTiles(k);
                    isWon();
                    if(isValidTile(tiles.indexOf(tiles.singleWhere((element) => element.key == k)))){
                      score--;
                    }
                  });
                }
              ),
            )
          ))
        );
      }
    }

    res[emptyTileIndex] = Flexible(child: Container(margin: const EdgeInsets.all(2)));

    return res;
  }

  void moveTiles(){
    for(int i=0; i < colNb * difficulty; i++) {
      var validTiles = getValidTiles();
      
      swapTiles( 
        validTiles[math.Random().nextInt(validTiles.length)].key!
      );
    }
  }

  void _showWinDialog(BuildContext context, int winscore){
    showDialog(
      context: context,
      builder: (BuildContext context){
        return AlertDialog(
          title: const Text('Congratulations !'),
          content: Column(
            mainAxisSize: MainAxisSize.min,
            children: <Widget>[
              Text('You won ! With a score of $winscore'),
              SizedBox(height: 20), // Add some space between text and image
              image,
            ],
          ),
          actions: <Widget>[
            TextButton(
              onPressed: () {
                Navigator.pop(context);
              },
              child: const Text('Close'),
            ),TextButton(
              onPressed: () { setState(() {
                  stopwatch = const Duration();
                  stopwatchTimer?.cancel();
                  stopwatchTimer = null;
                  score = difficulty * colNb;
             
                  image.image.evict();
                  image = new Image.network('https://picsum.photos/512');
                  tiles = buildImage();
                  moveTiles();
                });
                Navigator.pop(context);
                
              },
              child: const Text('Restart'),
            )
          ],
        );
      },
    );
  }

  void _showHintDialog(BuildContext context){
    showDialog(
      context: context,
      builder: (BuildContext context){
        return AlertDialog(
          title: const Text('Good luck !'),
          content: Column(
            mainAxisSize: MainAxisSize.min,
            children: <Widget>[
              Text('This is the image you are trying to recompose :'),
              SizedBox(height: 20), // Add some space between text and image
              image,
            ],
          ),
          actions: <Widget>[
            TextButton(
              onPressed: () {
                Navigator.pop(context);
              },
              child: const Text('Close'),
            ),
          ],
        );
      },
    );
  }


  bool isWon(){
    for(int i = 1; i < tiles.length; i++){

      if (tiles[i].key != initialOrder[i]){
        return false;
      }
    }

    //print("It's a win");
    stopwatch = const Duration();
    stopwatchTimer?.cancel();
    stopwatchTimer = null;

    final winscore = score - 1;
    _showWinDialog(context, winscore);

    return true;
  }

  bool isValidTile(int index) {
    return (index+1 == emptyTileIndex && index%colNb != colNb-1 ||
      index-1 == emptyTileIndex && index%colNb != 0 ||
      index-colNb == emptyTileIndex ||
      index+colNb == emptyTileIndex);
  }

  List<Widget> getValidTiles() {
    return tiles.where(
      (element) => isValidTile(tiles.indexOf(element))
    ).toList();
  }

  void swapTiles(Key k){
    Widget t = tiles.singleWhere((element) => element.key == k);
    int i = tiles.indexOf(t);

    // If tile is neighbor of empty tile => move
    if(isValidTile(i)){
      tiles[emptyTileIndex] = t;
      tiles[i] = Flexible(child: Container(margin: const EdgeInsets.all(2)));
      emptyTileIndex = i;
    }
  }

  Future getImageFromCamera() async {
    final pickedFile = await picker.pickImage(source: ImageSource.camera);

    if (pickedFile != null) {
      _image = File(pickedFile.path);
      image = Image.file(_image!);
      tiles = buildImage();
      moveTiles();
    }
  }

  Future getImage() async {
    final pickedFile = await picker.pickImage(source: ImageSource.gallery);

    if (pickedFile != null) {
      _image = File(pickedFile.path);
      image = Image.file(_image!);
      tiles = buildImage();
      moveTiles();
    }
  }

  @override
  Widget build(BuildContext context) {
    Widget timeWidget;
    
    timeWidget = stopwatchTimer != null ? Center(
        child: Text.rich(
          TextSpan(
            text: stopwatch.toString().substring(0, 9), // default text style
            style: const TextStyle()
          )
        )
      ) : const SizedBox.shrink();

    return Scaffold(
      appBar: AppBar(
        title: const Text('Jeu du Taquin !'),
        centerTitle: true,
      ),
      body: Column(children: [
        Text(
          'Score: $score',
          style: const TextStyle(fontSize: 20),
        ),
        const SizedBox(height: 5),
        Row(
          mainAxisAlignment: MainAxisAlignment.spaceEvenly,
          children: [
            ElevatedButton(
              onPressed: () {
                setState(() {
                  stopwatch = const Duration();
                  stopwatchTimer?.cancel();
                  stopwatchTimer = null;
                  score = difficulty * colNb;
             
                  image.image.evict();
                  image = new Image.network('https://picsum.photos/512');
                  tiles = buildImage();
                  moveTiles();
                });
              },
              child: const Text('Restart'),
            ),
            ElevatedButton(
              onPressed: () {
                setState(() {
                  
                  _showHintDialog(context);
                });
              },
              child: const Text('HINT'),
            ),
            ElevatedButton(
              onPressed: () {
                const d = Duration(milliseconds: 100);

                setState(() {
                  stopwatchTimer = Timer.periodic(d, (timer) {
                    setState(() {
                      stopwatch = stopwatch + d;
                    });
                  });
                });
              },
              child: const Text('Stopwatch'),
            ),
            // TODO : PLUS DE BOUTO
          
          ],
        ),
        const SizedBox(height: 5),
        Row(
          mainAxisAlignment: MainAxisAlignment.spaceEvenly,
          children: [
            ElevatedButton(
              onPressed: () {
                  getImage().then((_) {
                    setState(() {});
                  });
                  score = difficulty * colNb;
              },
              child: const Text('Image From Galeria'),
            ),
            ElevatedButton(
              onPressed: () {
                getImageFromCamera().then((_) {
                  setState(() {});
                });
                score = difficulty * colNb;
              },
              child: const Text('Image from camera'),
            ),
            
          
          
          ],
        ),
        const SizedBox(height: 5),
        Container(
          margin: const EdgeInsets.all(10.0),
          child: SizedBox(
            height: MediaQuery.of(context).size.height * 0.5,
            width: MediaQuery.of(context).size.height * 0.5,
            child: GridView.builder(
              gridDelegate: SliverGridDelegateWithFixedCrossAxisCount(
                crossAxisCount: colNb,
              ),
              itemCount: colNb * colNb,
              itemBuilder: (BuildContext context, int index) {
                return tiles.elementAt(index);
              }
            )
          )
        ),
        timeWidget
      ]),
    );
  }
}


class Tile {
  Image image;
  Alignment alignment;

  Tile({required this.image, required this.alignment});

  Widget croppedImageTile(double widthFactor) {

    return FittedBox(
      fit: BoxFit.fill,
      child: ClipRect(
        child: Container(
          child: Align(
            alignment: alignment,
            widthFactor: widthFactor,
            heightFactor: widthFactor,
            child: image,
          ),
        ),
      ),
    );
  }
}